#!/usr/bin/perl
#
# rndCiv.pl - make randomized FreeCiv rulesets
# Copyright (C) 2003, Billy Naylor <banjo@actrix.co.nz>
#
# Version 0.8.4
# 
# This program is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.  See <http://www.fsf.org/copyleft/gpl.txt>.
# 
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
# for more details.
#

use strict;

# -----------------------------------------------------------
# Usage

my $helptext = <<EOF;

$0: make randomized FreeCiv rulesets
Usage: $0 [OPTIONS] packagelist

Options:

    -v/-vv/-vvv  verboseness, more 'v's more verbose
    -s/--size    number of techs to have
    -o/--outdir  directory to write to
                 defaults to ./output/
    -d/--data    data directory
    -t/--techs   must have techs
                 ( use '-' to continue to package list )
    -h/--help    print this help & exit
    -l/--list    list known packages

    -p/-pp       prereq 'strictness'

EOF

# -----------------------------------------------------------
# Configuration

# this is a little unix specific
#srand (time ^ $$ ^ unpack "%L*", `ps axww | gzip`);
srand( time );

our $debug = 0;

# maxiumum number of techs freeciv allows
my $MAXTECHS = 199;

# How many techs
my $maxtechs = 90;

# data directory
my $datadir = './rulesets';

# output directory
my $outdir = './output';

# how strict
my $strictness = 0;

# -----------------------------------------------------------
# Parse args

my ( @pkgs, $techs, @musthave );

while ( @ARGV ) {
	$_ = shift;

	if ( m/^-+h(elp)?$/ ) {
		print $helptext;
		exit 1;
	}

	elsif ( m/^-(p+)$/ ) {
		$strictness = length $1;
		$strictness > 2 and $strictness = 2; 
		$debug and print "Setting Strictness level to $strictness\n";
	}

	elsif ( m/^-+s(ize)?$/ ) {
		die "No size value\n" unless scalar @ARGV;
		my $size = shift;
		die "Unable to have size '$size'\n" unless $size =~ /^\d+$/;
		$maxtechs = $size;
	}

	elsif ( m/^-+(v+)$/ ) {
		$debug = length $1;
		$debug and print "Setting Debug level to $debug\n";
	}

	elsif ( m/^-+t(echs)?$/ ) {
		while ( @ARGV ) {
			$_ = shift;
			last if ( m/^-+$/ );
			push @musthave, $_;
		}
	} 

	elsif ( m/^-+o(outdir)?$/ ) {
		die "No output directory\n" unless scalar @ARGV;
		$outdir = shift;
	}

	elsif ( m/^-+d(ata)?$/ ) {
		die "No data directory given\n" unless scalar @ARGV;
		$_ = shift;
		die "Unable to locate data dir '$_'\n"
		 unless ( -d $_ );
		$datadir = $_;
	}

	elsif ( m/^-+l(ist)?$/ ) {
		die "No data directory found\n" unless ( -d $datadir );
		print join ', ', map{ m/\.(\w+)$/; $1 } qx{ ls $datadir/techs.* };
		print "\n";
		exit 1;
	}

	elsif ( m/^-/ ) {
		next if m/^-+-+$/;
		die "Unknown option '$_'\n";
	}

	else { # packages
		die "Bad package name '$_'\n"
		unless m/^\w+$/;
		push @pkgs, $_;
	}
}

# -----------------------------------------------------------
# Parse tech packages


=pod $techs

  $techs->{
	$techname {
		{flags},
		{prereqs},
		{Real},
		{techlevel},
	},
  }

=cut

my ( $namedtechs, $pkgseen, $pastTop, $pkgcount ) = ( 'Package list:', {}, 0, 0 );
my $tmpPreReqs;
push @pkgs, '-';  # top of stack
my $techLevels;
while ( scalar @pkgs ) {
	my $pkg = shift @pkgs;

	# are we past the packages listed on the command line ?
	if ( $pkg =~ m/^\-$/ ) {
		$pastTop = 1;
		next;	
	}

	my $filename = "$datadir/techs.$pkg";
	die "Unknown package '$pkg'\n"
	  unless ( -e $filename );
	next if ( defined $pkgseen->{$pkg} );
	$pkgseen->{$pkg} = ++$pkgcount;
	$namedtechs .= " $pkg";

	open( TECHS, "< $filename" )
	  or die "Unable to open package $pkg '$filename':$!\n";

	my $linecount = 0;
    my $prevname  = '';
	my $currentTechLevel = 100;  # default is set very high
	for (<TECHS>) {
		$linecount++;
		next if m/^#/;      # ignore comments ( hash in column 1 )
		next if m/^\s*$/;   # and blank lines
		my $errhelp = "$_ [$filename : $linecount]";
		chomp;

		# grab commands
		if ( m/^\+\s+(\w+)\s+(.*?)\s*$/ ) {
			my( $cmd, $arg ) = ( $1, $2 );

			if ( $cmd eq 'read' ) {
				my @choices = &shuffle( split /\s*\|\s*/, $arg );
				( scalar @choices ) or
				 warn "No options listed $errhelp\n" and next;
				push @pkgs, shift @choices;
			}

			elsif ( $cmd eq 'level' ) {
				if ( $arg !~ m/^\d+$/ ) {
					warn "Bad tech level '$arg' $errhelp\n";
				}
				else {
					$currentTechLevel = $arg;
				}
			}

			elsif ( $cmd eq 'choose' ) {
#				next if $pastTop;
				my @choices = &shuffle( split /\s*\|\s*/, $arg );
				( scalar @choices ) or
				 warn "No options listed $errhelp\n" and next;
				push @musthave, shift @choices;
			}

			else {
				warn "Bad command: $errhelp\n";
			}
			next;
		}

		# Speech
		if ( m/^\s*("[^"]+")\s*$/ ) {
			$prevname or
			 die "Invalid speech $errhelp\n";
			$techs->{$prevname}{Speech} = $1;
			next;
		}

		# added to Other tech
		if ( m/^\s*->\s*(.*?)$/ ) {
			$debug > 1 and print "pushed: $prevname into $1\n";
			$prevname or
			 die "Invalid line $errhelp\n";
			map { push @{ $tmpPreReqs->{$_} }, $prevname }
			 split /\s*,\s*/, $1;
			next;
		}

		# grab techs
		next unless m/
			^(\w+.*?)		# name
			\:\s+			# name delim
			(\((.*?)\))?	# possible flags
			\s*(.*?)\s*$	# requirements
		/x;
		my ( $name, $flags, @reqs ) = ( $1, $3, split( /\s*,\s*/, $4 ) );
		$flags ||= '';
		$prevname = $name;

		$techs->{$name}{flags} = $flags
		 unless ( exists $techs->{$name}{flags} and rand(1) < 0.2 );
		$techs->{$name}{techlevel} = $currentTechLevel;
		push @{ $tmpPreReqs->{$name} }, @reqs;

		push @{ $techLevels->{$currentTechLevel} }, $name;

	}
	close TECHS;
}

for my $name ( keys %{ $techs } ) {
	next unless exists $tmpPreReqs->{$name};

	if ( scalar  @{ $tmpPreReqs->{$name} } < 2 ) {
		delete $techs->{$name};
		print "Tech $name dropped with too few prereqs\n";
		next;
	};

	push @{ $techs->{$name}{prereqs} },
	 grep { exists  $techs->{$_} }
	  &shuffle( @{ $tmpPreReqs->{$name} } );
}
undef $tmpPreReqs;

my @tek = keys %{ $techs };
die "No tech packages loaded\n"
 unless ( scalar @tek );

$maxtechs = ( scalar @tek ) if ( scalar @tek < $maxtechs );
$maxtechs = $MAXTECHS if ( $maxtechs > $MAXTECHS );


# -----------------------------------------------------------
# Build techtree

=pod $techtree

 $techtree->{
	'None' => 1,
	'Name' => $name,
 }

=cut

my $techtree  = { None => { req1 => undef, req2 => undef, node => rndCiv::node->new( 'None' ) } };
my $techcount = 0;

# first get techs from command line, our musthaves
for my $tech ( sort { $techs->{$a}{techlevel} <=> $techs->{$b}{techlevel}  }  @musthave ) {
	&buildown( $tech )
	 or $debug and print "Unable to build tree down for tech '$tech'\n";
}

# second build up by techlevel
for my $tl ( sort { $a <=> $b } keys %$techLevels ) {
	last if ( $techcount > $maxtechs );
	next unless defined $techLevels->{$tl};
	my $tech = pop @{ $techLevels->{$tl} };
	&buildown( $tech )
	 or $debug and print "Unable to build tree down for tech '$tech'\n";
}

# lastly if we still havn't enough, just start to grab techs at random
for my $tech ( &shuffle( @tek ) ) {
	last if ( $techcount > $maxtechs );
	&buildown( $tech )
	 or $debug and print "Unable to build tree down for tech '$tech'\n";
}
undef @pkgs;

# -----------------------------------------------------------
# Write techs.ruleset & techtree.dot

unless ( -d $outdir ) {
	mkdir $outdir
	 or die "Unable to make output directory '$outdir':$!\n";
}

open ( FILE, "> $outdir/techs.ruleset" )
 or die "Unable to open '$outdir/techs.ruleset' for writing:$!\n";

print FILE <<EOF;
; created by rndCiv.pl
; total $techcount techs

[datafile]
description="Randomized technology data for Freeciv"
options="1.9"

EOF

open ( DOT, "> $outdir/techtree.dot" )
 or die "Unable to open '$outdir/techtree.dot' for writing:$!\n";

print DOT <<EOF;
digraph techtree {
	ordering=out;
	ratio=auto;
	node [shape=box];
	rankdir=LR;
	mclimit=2.0;

EOF

my $summary = {
	'None' => {
		pre       => [],
		units     => [],
		buildings => [],
		govs      => [],
		steps     => 0,
	},
};

# dunno where this comes from?
delete $techtree->{''};

for my $name ( sort keys %{ $techtree } ) {
	next if ( $name eq 'None' or $name eq '' );
	my ( $req1, $req2, $flags ) = (
		$techtree->{$name}{req1},
		$techtree->{$name}{req2},
		$techs->{$name}{flags}
	);

	my $advance = lc( $name );
	$advance =~ s/ /_/g;

	( $req1, $req2 ) = ( $req2, $req1 )
	 if ( $req1 eq 'None' and $req2 ne 'None' );

	$flags ||= '""';

	print FILE "[advance_$advance]\n"
	 . "name        = _(\"$name\")\n"
	 . "req1        = \"$req1\"\n"
	 . "req2        = \"$req2\"\n"
	 . "flags       = $flags\n"
	 . "graphic     = \"a.alphabet\"\n"
#	 . "graphic     = \"a.$advance\"\n"
	 . "graphic_alt = \"-\"\n\n";

	print DOT "\t\"$req1\" -> \"$name\";\n"
	 . "\t\"$req2\" -> \"$name\";\n";

	$summary->{$name}{pre}       = [ $req1, $req2 ];
	$summary->{$name}{units}     = [];
	$summary->{$name}{buildings} = [];
	$summary->{$name}{govs}      = [];
	$summary->{$name}{cities}    = [];
	$summary->{$name}{steps}     = $techtree->{$name}{steps};
	push @{ $summary->{$req1}{allows} }, $name;
	push @{ $summary->{$req2}{allows} }, $name;
}

close FILE;
print DOT "}\n";
close DOT;

# -----------------------------------------------------------
# set up lookups lists

my %tagz = qw{
	T techs
	U units
	B buildings
	G governments
	C cities
	E effects
	P Packages
};

my $TAGS = 'TUBGCEP';  # Techs, Units, Buildings, Govs, Cites, Effects

my $got = {
	techs       => {},
	units       => {},
	buildings   => {},
	governments => {},
	cities      => {},
	effects     => {},
	packages    => {},
};

map { $got->{packages}{$_} = 1 } keys %{ $pkgseen };
map { $got->{techs}{$_} = 1    } keys %{ $techtree };
undef $techtree;

# -----------------------------------------------------------
# read in governments, buildings and units data files
my @dataList;
{
	my @pkgList = keys %{ $pkgseen };

	for my $type ( qw{ units buildings governments cities effects } ) {
		for my $p ( @pkgList ) { 
			my $filename = "$datadir/$type.$p";
			$debug > 1 and print "Reading package '$filename'\n";

			unless( open( FILE, "< $filename" ) ) {
				$debug > 2 and print "units ruleset '$filename' not found\n";
				next;
			}
	
			my $buf = '';
			sysread( FILE, $buf, -s $filename );
			close FILE;
			$buf =~ s/^\s*\#.*?$//mg;

			$buf =~ s/^.*?\nname\s+=/name =/s; # chop off the intro
			for my $block ( split /^name\s+=/m, $buf ) {
				next unless $block =~ m/\w/;
				my $s = rndCiv::Data->new( $type, "name = $block", $filename );

				# 60% chance of keeping old when name collision
				next if (
					exists $got->{$type}{ $s->{name} }
					and defined $got->{$type}{ $s->{name} }
					and  rand(1) < 0.6
				);

				$got->{$type}{ $s->{name} } = $s;
				push @dataList, $s;
			}
		}
	}
}

# -----------------------------------------------------------
# expand out Effects - they have ways of wrapping themselves up

{
	# handle '&' cases
	for my $s ( @dataList ) {

		for my $effect ( grep /^(effect|reqs)/, keys %{ $s->{fields} } ) {
			my $e = $s->{fields}{$effect};
			next unless $e =~ m/\&/;

			die "Bad field in $s->{name}\n$e\n"
			 . " '&' not allowed in '{ }'\n"
			 if ( $e =~ m/\{[^\}]+\&/ );

			die "Bad field in $s->{name}\n$e\n"
			 .  "$s->{type} cannot use '&' operator\n"
			 unless $s->{type} =~ m/^(buildings|governments|effects)$/;

			while ( $e =~ m/^(\s+.*?\<[$TAGS]\s+([^\>]+?\&[^\>\|]+)(\s*\|?[^\>]*)?\s*\>.*?)$/m ) {
				my ( $line, $block ) = ( $1, $2 );
				my @tags =
				my $buf = '';
				for my $t ( split /\s*\&\s*/, $block ) {
					my $newline = $line;
					$newline =~ s/\Q$block/$t /;
					$buf .= "\n$newline";
				}
				$e =~ s/\Q$line/$buf/;
				$e =~ s/\n\n/\n/g;
			}
			$s->{fields}{$effect} = $e;
		}
	}

	# handle '%' cases
	for my $s ( @dataList ) {

		for my $effect ( grep /^effect/, keys %{ $s->{fields} } ) {
			my $data = $s->{fields}{$effect};
			next unless $data =~ m/\%/m;
			die "Bad field in $s->{name}\n$data\n"
			 .  "$s->{type} cannot use '%' operator\n"
			 unless $s->{type} =~ m/^(buildings|governments|effects)$/;

			my @newEffects;
			my @effects = ( $data );
			while ( scalar @effects > 0 ) {
				my $e = shift @effects;
				unless ( $e =~ m/\%/m ) {
					push @newEffects, $e;
					next;
				}

				my ( $tag, $options );
				# in <>
				if ( $e =~ m/(\<[$TAGS]\s+([^\>]+?\%[^\>\|]+?)\s+(\|\s*\w*\s*)?\s*\>)/m ) {
					( $tag, $options ) = ( $1, $2 );
				}
				# in {}
				elsif ( $e =~ m/(\{\s*([^\}]+?\%[^\}]+?)\s*\})/ ) {
					( $tag, $options ) = ( $1, $2 );
				}
				else {
					die "Badly formatted '%' line in $s->{type} $s->{name}\n$e\n:";
				}
				my @tags = split /\s*\%\s*/, $options;

				for my $t ( @tags ) {
					my $newEffect = $e;
					$newEffect =~ s/\Q$options/$t/;
					$newEffect =~ s/\{\s*(\w+?)\s*\}/$1/g;
					push @effects, $newEffect;
				}
			}
			delete $s->{fields}{$effect};
			my $count = 0;
			for my $ne ( @newEffects ) {
				$count++;
				$s->{fields}{"${effect}_$count"} = $ne;
			}
		}
	}
}

# -----------------------------------------------------------
# check dependencies are ok, mix and match till we get it sorted
# loops until no more good things found
# uses the 'level' field as a guide
#
# the effects file means that Buildings, Units and Governments
# have arbitrary dependencies upon each other, maybe even circular deps
{

	# step 1 - loop through our list removing items that can never be made
	# and looking for conflicts
	my $gotone = 1;
	$debug > 1 and print "Entering remove dependencies loop\n";
	while ( $gotone ) {
		$gotone = 0; 
		die "Ran out of items" if scalar @dataList == 0;

		my ( @good, @conflicted );
		for my $s ( @dataList ) {
			if ( $s->isok ) {
				if ( $s->hasConflicts ) { push @conflicted, $s }
				else { push @good, $s }
				next;
			}
			$gotone = 1;
			$debug > 2 and print "\tremoved $s->{type} $s->{name}\n";
			delete $got->{ $s->{type} }{ $s->{name} };
		}

		# remove rnd conflicted
		if ( not $gotone and scalar @conflicted ) {
			@conflicted = &shuffle( @conflicted );
			my $bad = pop @conflicted;
			$gotone = 1;
			$debug > 2 and print "\tremoved conflicted $bad->{type} $bad->{name}\n";
			delete $got->{ $bad->{type} }{ $bad->{name} };
		}

		@dataList = ( @good, @conflicted );
	}

	# step 2 - do basic tech reqs, so we can get techlevels 
	$debug > 1 and print "Entering tech dependencies loop\n";
	for my $s ( @dataList ) {
		$s->techDepends;
	}

	# step 3 - everything should be good (fingers crossed)
	$debug > 1 and print "Entering check dependencies loop\n";
	for my $s ( @dataList ) {
		$debug > 2 and print "Making $s->{name}\n";
		$s->make;
	}

	undef @dataList;	
}

my $Effects;

# ===========================================================

$debug and print "Writing governments.ruleset\n";
{

	my $output = <<EOF;
; created by rndCiv.pl

[datafile]
description="Randomized governments data for Freeciv"
options="1.9"

[governments]
when_anarchy="Anarchy"

EOF

	for my $s ( map{ $got->{governments}{$_} } sort keys %{ $got->{governments} } ) {
		if ( $s->{fields}{reqs} =~ m/Tech\s*,\s*(.+?)\s*,/ ) {
			my $req = $1;
			( exists $summary->{$req} )
			 and push @{ $summary->{$req}{govs} }, $s->{name};
		}
		$output .= "; -------------------------------------------------------\n"
		 . "[government_$s->{lc_name}]\n\n"
		 . $s->getField( 'name', 'n', 'r' )
		 . $s->getReqs()
		 . $s->getResource()
		 . $s->getField( 'ai_better', 's', 'c' )
		 . $s->getRulerTitles()
		 . $s->getHelpText;

		$Effects .= $s->getEffects();
	}

	# print to file
	open ( FILE, "> $outdir/governments.ruleset" )
	 or die "Unable to open '$outdir/governments.ruleset' for writing:$!\n";
	print FILE $output;
	close FILE;
}

# -----------------------------------------------------------

$debug and print "Writing buildings.ruleset\n";
{

	my $output = <<EOF;
; created by rndCiv.pl

[datafile]
description="Randomized buildings data for Freeciv"
options="1.10.1"

EOF

	for my $s ( map{ $got->{buildings}{$_} } sort keys %{ $got->{buildings} } ) {
		if ( $s->{fields}{reqs} =~ m/Tech\s*,\s*(.+?)\s*,\s*Player/ ) {
			my $req = $1;
			( exists $summary->{$req} )
			 and push @{ $summary->{$req}{buildings} }, $s->{name};
		}
		$output .= "[building_$s->{lc_name}]\n"
		 . $s->getField( 'name', 'n', 'r' )
		 . $s->getField( 'genus', 's', 'r' )
		 . $s->getReqs()
		 . $s->getResource()
		 . $s->getField( 'obsolete_by', 's', 'n' )
		 . $s->getField( 'replaced_by', 's', 'c' )
		 . $s->getField( 'build_cost', 'd', 'r' )
		 . $s->getField( 'upkeep', 'd', 'z' )
		 . $s->getField( 'sabotage', 'd', 'z' )
		 . $s->getHelpText;

		$Effects .= $s->getEffects();
	}

	# print to file
	open ( FILE, "> $outdir/buildings.ruleset" )
	 or die "Unable to open '$outdir/buildings.ruleset' for writing:$!\n";
	print FILE $output;
	close FILE;
}

# -----------------------------------------------------------

$debug and print "Writing units.ruleset\n";
{

	my $output = <<EOF;
; created by rndCiv.pl

[datafile]
description="Default unit_type data for Freeciv"
options="1.9"

[veteran_system]
veteran_names = _("green"), _("veteran"), _("hardened"), _("elite")
veteran_raise_chance = 50, 33, 20, 0
veteran_power_fact = 100, 150, 175, 200
veteran_work_raise_chance = 5, 3, 1, 0
veteran_highseas_loss_pct = 50, 25, 5, 0
veteran_move_bonus = 0, 0, 0, 0

EOF

	for my $s ( map{ $got->{units}{$_} }
	 sort { $got->{units}{$a}{level} <=> $got->{units}{$b}{level} }
	 keys %{ $got->{units} } ) {

		my $req = $s->{fields}{tech_req};
		( exists $summary->{$req} )
		 and push @{ $summary->{$req}{units} }, $s->{name};

		$output .= "[unit_$s->{lc_name}]\n"
		 . $s->getField( 'name', 'n', 'r' )
		 . $s->getField( 'class', 's', 'r' )
		 . $s->getField( 'tech_req', 's', 'n' )
		 . $s->getField( 'impr_req', 's', 'c' )
		 . $s->getField( 'gov_req', 's', 'c' )
		 . $s->getField( 'obsolete_by', 's', 'n' )
		 . $s->getResource()
		 . $s->getField( 'build_cost', 'd', 'r' )
		 . $s->getField( 'pop_cost', 'd', 'z' )
		 . $s->getUnitCombat()
		 . $s->getUnitMove()
		 . $s->getField( 'vision', 'd', 'r', 'vision_radius_sq' )
		 . $s->getField( 'transport', 'd', 'z', 'transport_cap' )
		 . $s->getUnitUpkeep()
		 . $s->layoutWords( 'flags' )
		 . $s->layoutWords( 'roles' )
		 . $s->getHelpText;

		$Effects .= $s->getEffects();
	}

	# print to file
	open ( FILE, "> $outdir/units.ruleset" )
	 or die "Unable to open '$outdir/units.ruleset' for writing:$!\n";
	print FILE $output;
	close FILE;
}

# -----------------------------------------------------------

$debug and print "Writing cities.ruleset\n";
{
	my $output = <<EOF;
; created by rndCiv.pl

[datafile]
description="Cities data for Freeciv"
options="1.9"

[specialist_elvis]
name            = _("elvis")
short_name      = _("?Elvis:E")
reqs            =
    { "type", "name", "range"
; Nothing
    }

[specialist_scientist]
name            = _("scientist")
short_name      = _("?Scientist:S")
reqs            =
    { "type", "name", "range"
      "MinSize", "5", "City"
    }

[specialist_taxman]
name            = _("taxman")
short_name      = _("?Taxman:T")
reqs            =
    { "type", "name", "range"
      "MinSize", "5", "City"
    }

[parameters]
add_to_size_limit  = 8
angry_citizens = 1              ; set to zero to disable angry citizens
celebrate_size_limit = 3        ; cities >= can celebrate
changable_tax = 1               ; set to zero to disallow changing of tax rates

;forced_science = 0             ; set these fields when changeable_tax is turned off
;forced_luxury = 100
;forced_gold = 0

EOF

	for my $s ( map{ $got->{cities}{$_} } sort keys %{ $got->{cities} } ) {

		my $req = 'None';
		if ( exists $s->{fields}{reqs} and defined $s->{fields}{reqs}
		 and $s->{fields}{reqs} =~ m/Tech\s*\,\s*(.+?)\,/ ) {
			$req = $1;
		}
		( defined $req and exists $summary->{$req} )
		 and push @{ $summary->{$req}{cities} }, $s->{name};

		$output .= "[citystyle_$s->{lc_name}]\n"
		 . $s->getField( 'name', 'n', 'r' )	
		 . $s->getResource()
		 . $s->getReqs()
		 . $s->getField( 'replaced_by', 's', '' )
		 . "\n";
	}

	# write the cities.ruleset
	open (FILE, "> $outdir/cities.ruleset" )
	 or die "Unable to open '$outdir/cities.ruleset' for writing:$!\n";
	print FILE $output;
	close FILE;
}

# -----------------------------------------------------------

my $summaryEffects;

$debug and print "Writing effects.ruleset\n";
{
	for my $s ( map{ $got->{effects}{$_} } keys %{ $got->{effects} } ) {
		$Effects .= $s->getEffects();
		$summaryEffects .= "$s->{name}, ";
	}

	my $output = <<EOF;
; created by rndCiv.pl

[datafile]
description="Default effects data for Freeciv"
options="1.0"

; /* <-- avoid gettext warnings
; */ <-- avoid gettext warnings

$Effects

EOF

	open(EFFECTS, "> $outdir/effects.ruleset" )
	 or die "Unable to open '$outdir/effects.ruleset':$!\n";
	print EFFECTS $output;
	close EFFECTS
}
 
# -----------------------------------------------------------
# write summary

open ( FILE, "> $outdir/ruleset.summary" )
 or die "Unable to open '$outdir/ruleset.summary' for writing:$!\n";

print FILE "Summary of rndCiv ruleset"
 . "\nDate: " . scalar localtime( time )
 . "\n$techcount techs"
 . "\n$namedtechs\n";

my $area = {};

for my $t ( sort keys %{ $summary } ) {
	# bogus hack to fix bad summary field - dunno why it's there
	next unless $t;

	my $steps = $summary->{$t}{steps};
	print FILE "\n$t [$steps](" . join ', ', @{ $summary->{$t}{pre} };
	print FILE ")";
	$area->{techs}{$steps} .= "$t, ";

	if ( exists $summary->{$t}{allows} ) {
		scalar @{ $summary->{$t}{allows} } and
		print FILE "\n allows: ", join ', ', @{ $summary->{$t}{allows} };
	}

	if ( scalar @{ $summary->{$t}{units} } ) {
		my $buf = join ', ', @{ $summary->{$t}{units} };
		print FILE "\n units: $buf";
		$area->{units}{$steps} .= "$buf, ";
	}

	if ( scalar @{ $summary->{$t}{buildings} } ) {
		my $buf = join ', ', @{ $summary->{$t}{buildings} };
		print FILE "\n buildings: $buf";
		$area->{buildings}{$steps} .= "$buf, ";
	}

	if ( scalar @{ $summary->{$t}{govs} } ) {
		my $buf = join ', ', @{ $summary->{$t}{govs} };
		print FILE "\n governments: $buf";
		$area->{govs}{$steps} .= "$buf, ";
	}

	if ( scalar @{ $summary->{$t}{cities} } ) {
		my $buf = join ', ', @{ $summary->{$t}{cities} };
		print FILE "\n cities: $buf";
		$area->{cities}{$steps} .= "$buf, ";
	}

}

print FILE "\n\n ------------------------------------\n";

print FILE "\nTechs: \n";
for my $k ( sort { int $a <=> int $b } keys %{ $area->{techs} } ) {
        print FILE "$k: " . $area->{techs}{$k} . "\n";
}

print FILE "\nUnits: \n";
for my $k ( sort { int $a <=> int $b } keys %{ $area->{units} } ) {
        print FILE "$k: " . $area->{units}{$k} . "\n";
}

print FILE "\nBuildings: \n";
for my $k ( sort { int $a <=> int $b } keys %{ $area->{buildings} } ) {
        print FILE "$k: " . $area->{buildings}{$k} . "\n";
}

print FILE "\nGovernments: \n";
for my $k ( sort { int $a <=> int $b } keys %{ $area->{govs} } ) {
        print FILE "$k: " . $area->{govs}{$k} . "\n";
}

print FILE "\nCities: \n";
for my $k ( sort { int $a <=> int $b } keys %{ $area->{cities} } ) {
        print FILE "$k: " . $area->{cities}{$k} . "\n";
}

print FILE "\nEffects: \n$summaryEffects\n\n";

print FILE "\n\n";
close FILE;

# -----------------------------------------------------------
# thankyouverymuchgoodnight
exit 1;

# -----------------------------------------------------------
# subs

# -----------------------------------------------------------
# The Fisher Yates Shuffle from the Perl Cookbook
# 
sub shuffle {
	my @array = @_;
	for my $i ( 0..$#array ) {
		my $j = int rand( $i + 1 );
		next if $i == $j;
		( $array[$i], $array[$j] ) = ( $array[$j], $array[$i] );
	}
	return wantarray ?  @array : \@array;
}

# -----------------------------------------------------------
# begin to build down to 'None' from a given tech
#
sub buildown {
	my $tech = shift;
	return 1 if ( exists $techtree->{$tech} );
	$debug and print "Building down from $tech\n";

	my ( $tree, %tekList ) = &buildnodesdown( $tech, () )
	 or return 0;

	&readtree( $tree )
	 or return 0;

	return 1;
}

# -----------------------------------------------------------
# create a tech tree down to 'None'
# might redefine working techs ??
#
sub buildnodesdown {
	my ( $name, %tekList ) = @_;

	if ( exists $techtree->{$name} ) {
		$debug > 2 and print "'$name' is in techtree\n";
		return $techtree->{$name}{node};
	}

	unless ( exists $techs->{$name} ) {
		#$debug > 1 and
		print "Unknown tech '$name' called\n";
		return undef;
	}

	# already found 
	if ( exists $tekList{$name} ) {
		$debug > 2 and print "already have $name\n";
		return $tekList{$name}{node};
	}
	$tekList{$name}{node} = undef;

	# choose an available prereq
	my ( $pre1, $pre1name, $hold, $holdname );
	for my $p ( &shuffle( @{ $techs->{$name}{prereqs} } ) ) {
		$debug > 1 and print "$name: Trying prereq $p\n";

		# deal with the easy 'None'
		if ( $p eq 'None' ) {
			if ( defined $pre1name ) {
				next if defined $holdname;
				$holdname = 'None';
				$hold = rndCiv::node->new( 'None' );
				next;
			}
			else {
				$pre1name = 'None';
				$pre1 = rndCiv::node->new( $pre1name );
				next;
			}
		}

		# recurse !
		my ( $ret, %newTL ) = &buildnodesdown( $p, %tekList );

		unless ( defined $ret ) {
			$debug > 2 and print "No path found for '$p'\n";
			next;
		}
		next if $ret->contains( $name );

		unless ( $pre1 ) {
			$pre1 = $ret;
			$pre1name = ( ref $pre1 ) ? $pre1->{name} : $pre1;
			if ( $pre1name eq 'None' ) {
				$hold = rndCiv::node->new( 'None' );
				last;
			}
			next;
		}
		my $pre2 = $ret;

		my $pre2name = ( ref $pre2 ) ? $pre2->{name} : $pre2;
		next if ( $pre2name eq $pre1name );

		unless ( $strictness == 0 or $pre1name eq 'None' or $pre2name eq 'None' ) {
			my $score = 2;
			$score-- if (
				$pre1->{req1}{name} eq $pre2name 
			 or $pre1->{req2}{name} eq $pre2name );
			$score-- if (
				$pre2->{req1}{name} eq $pre1name 
			 or $pre2->{req2}{name} eq $pre1name );
			next if ( $score != 2 or $score < $strictness );
		}

		if ( $pre2name eq "None" ) {
			#next if ( $holdname ne 'None' );
			( $hold,  $holdname ) = ( $pre2, $pre2name );
			next;
		}
		
		( $hold,  $holdname ) = ( $pre2, $pre2name );
		if ( $pre1->contains( $pre1name )
		  or $pre2->contains( $pre2name ) ) {
			$debug > 2 and print "Holding $pre2name\n";
			next;
		}

		last;
	}

	unless ( $pre1 and $hold ) {
		$debug > 1 and print "Failed to find 2 prereqs for tech '$name'\n";
		return undef
	}

	$debug > 2 and print "New node for $name( $pre1name, $holdname )\n";
	my $node = rndCiv::node->new( $name );
	$node->{req1} = $pre1;
	$node->{req2} = $hold;
	return ( $node, %tekList );
}

# -----------------------------------------------------------
# parse sucessful tree into $techtree
#
sub readtree {
	my $node = shift;
	my $name = ( ref $node ) ? $node->{name} : $node;
	return $name
	  if ( exists $techtree->{$name} );

	$techtree->{$name} = {};
	$techtree->{$name}{node} = $node;
	my @steps;
	for my $r ( 'req1', 'req2' ) {
		$techtree->{$name}{$r} = ( ref $node->{$r} ) ?
		 &readtree( $node->{$r} ) : $node->{$r};
		push @steps, $techtree->{ $techtree->{$name}{$r} }{steps};
	}
	$techtree->{$name}{steps} =
	 ( $steps[0] > $steps[1] ) ?
	  $steps[0] + 1 : $steps[1] + 1 ;

	$techcount++;
	return $name;
}

# -----------------------------------------------------------
# object class for nodes

package rndCiv::node;

sub new {
	my ( $class, $name ) = @_;
	return bless {
		name  => $name,
		req1  => undef,
		req2  => undef,
		steps => 0,
	}, $class;
}

sub contains {
	my ( $self, $name ) = @_;
	return 0 if $self->{name} eq 'None';
	return 1 if $self->{req1}->contains( $name );	
	return 1 if $self->{req2}->contains( $name );	
	return 0;
}

# -----------------------------------------------------------
# class for Buildings, Units, Cities and Govs
#
package rndCiv::Data;

sub new {
	my ( $class, $type, $buf, $filename ) = @_;

	my $s = bless {
		filename  => $filename,
		name      => '',
		lc_name   => '',
		level     => 0,
		group     => '',
		conflicts => [],
		type      => $type,
		fields    => {},
	}, $class;

	# get name
	if ( $buf =~ s/^name\s+=\s*(.+?)\s*$//m ) {
		$s->{name} = $1;
		$s->{fields}{name} = $1;
		$s->{lc_name} = lc $s->{name};
		$s->{lc_name} =~ s/ +/_/g;
		$s->{lc_name} =~ s/\W//g;
	}
	$debug > 1 and print "Adding $type $s->{name}\n";

	# get conflicts
	if ( $buf =~ s/^conflicts\s+=\s+(.+?)\s*$//m ) {
		@{ $s->{conflicts} } = grep /^[$TAGS]/, split /\s*\>\s*\<\s*|\s*\<\s*|\s*\>\s*/, $1;
	}

	# get helptext
	if ( $buf =~ s/^helptext\s+=\s+\"(.+?)\"\s*\n//ms ) {
		$s->{fields}{helptext} = $1;
	}

	# get level - only hardcoded for units
	if ( $buf =~ s/^level\s+=\s+(\d+)\s*$//m ) {
		$s->{level} = $1;
	}

	# get group(s)
	if ( $buf =~ s/^group\s+=\s+(.+?)\s*$//m ) {
		$s->{group} = join '|', split /\s*\,\s*/, $1;
	}

	# get reqs
	if ( $buf =~ s/^reqs\s+=\s*(.+(\n\s+and\s+.+?)*)\s*$//m ) {
		$s->{fields}{reqs} = $1;
	}

	# get effects clauses
	my $effcount = 0;
	while ( $buf =~ m/^effect\s+=/m ) {
		$effcount++;
		$buf =~ s/^(effect\s+=\s+.+\n(\s+if\s+.+(\n\s+and\s+.+)*(\n\s+unless\s+.+)*)?)//m;
		$s->{fields}{"effect_$effcount"} = $1;
	}

	# get other fields
	while ( $buf =~ s/^(\w+)\s+=\s+(.+?)?\s*$//m ) {
		$s->{fields}{$1} = $2;
	}

	return $s;
}

# checks to see whether object can fulfill it's dependencies
sub isok {
	my $s = shift;

	$debug > 1 and print "isok? $s->{name}\n";
	for my $f ( keys %{ $s->{fields} } ) {
		next if $f eq 'helptext';
		next if $f eq 'obsolete_by';	
		next unless $s->{fields}{$f} =~ m/\</;

		for ( split '<', $s->{fields}{$f} ) {
			next unless /([$TAGS])\s+([^\<\>]+?)\s*\>/;
			my ( $type, $options ) = ( $tagz{$1}, $2 );	

			my $gotone = 0;
			for my $opt ( split /\s*\|\s*/, $options ) {
				if ( $opt =~ m/^(None|cancel|skip|later|\-)$/ ) {
					$gotone = 1;
					last;
				}
				elsif ( exists $got->{$type}{$opt} ) {
					$gotone = 1;
					last;
				}
			}
			return 0 unless $gotone;
		}
	}

	return 1;
}

sub hasConflicts {
	my $s = shift;
	return 0 unless ( scalar @{ $s->{conflicts} } );

	for ( @{ $s->{conflicts} } ) {
		m/^([EGUCTB])\s+(.+?)\s*$/
		 or die "Bad conflict '<$_>' in $s->{type} $s->{name} (from $s->{filename})\n";
		return 1 if ( exists $got->{ $tagz{$1} }{ $2 } );
	}
	return 0;
}

sub techDepends {
	my $s = shift;
	next if $s->{type} =~ m/^(units|effects)$/;

	$s->makeField( 'reqs' );
	$s->{fields}{reqs} =~ m/Tech\s*,\s*(.*?)\s*,\s*Player/;
	$s->{level} = $summary->{ $1 }{steps};
}

# object fulfils it's dependencies
sub make {
	my $s = shift;

	for my $f ( keys %{ $s->{fields} } ) {
		next if ( $f eq 'helptext' );
		next unless $s->{fields}{$f} =~ m/</;  # shortcircuit
		$s->makeField( $f );
	}
}

sub makeField {
	my ( $s, $f ) = @_;

	while ( $s->{fields}{$f} =~ m/(\<([$TAGS])\s+(.+?)\s*\>)/ ) {
		my ( $tag, $type, $options ) = ( $1, $tagz{$2}, $3 );

		my $gotone = 0;
		foreach my $opt ( split /\s*\|\s*/, $options ) {
			if ( $opt eq 'None' ) {
				$gotone = 1;
				$s->{fields}{$f} =~ s/\Q$tag/None/;
			}
			elsif ( $opt eq 'cancel' ) {
				$gotone = 1;
				$s->{fields}{$f} = '';
			}
			elsif ( $opt eq '-' ) {
				$gotone = 1;
				$s->{fields}{$f} =~ s/\Q$tag/-/;
			}
			elsif ( $opt eq 'skip' ) {
				$gotone = 1;
				$s->{fields}{$f} =~ s/^.*?\Q$tag\E.*?$//m;
			}
			elsif ( $opt eq 'later' ) {
				my $later = $s->getLater() or next;	
				$gotone = 1;
				$s->{fields}{$f} =~ s/\Q$tag/$later/;
			}
			elsif ( exists $got->{$type}{$opt} ) {
				$gotone = 1;
				$s->{fields}{$f} =~ s/\Q$tag/$opt/;
			}
			last if $gotone;
		}

		unless ( $gotone ) {
			if ( $f eq 'obsolete_by' ) { $s->{fields}{$f} =~ s/\Q$tag/None/; }
			else {	die "Failed to make $s->{type} $s->{name} (field $f) (from $s->{filename})\n" }
		}
	}
}

# handle the 'later' option
sub getLater {
	my $s = shift;
	return undef unless $s->{group};

	my ( $type, $re, $lev ) = ( $got->{ $s->{type} }, $s->{group}, $s->{level} );
	my @list =
	 sort { $type->{$a}{level} <=> $type->{$b}{level} }
	 grep { $type->{ $_ }{group} =~ m/^($re)$/ }
	 grep { $type->{ $_ }{level} > $lev }
	 keys %{ $type };

	return undef unless scalar @list;
	return shift @list;
}

# returns formatted effects
sub getEffects {
	my $s = shift;

	my $buf;
	my $count = 0;
	for my $f ( grep /^effect/, keys %{ $s->{fields} } ) {
		# some effects were whacked by make()
		next unless $s->{fields}{$f};

		my $effect = $s->{fields}{$f};
		$effect =~ s/^effect\s*=\s*(.*?)\s*\n//
		 or die "$s->{name} (from $s->{filename}) has Bad effect head: '$effect'\n";
		my $head = $1;

		for my $h ( split /\s+\&\s+/, $head ) {
			my $e = $effect . "\n";

			# head line
			$h =~ m/^(\w+)\s*,\s*(-?\d+)\s*(,\s*S\s*)?$/
			 or die "$s->{type} $s->{name}\[$f] (from $s->{filename}) has bad effect\n'$h'";
			my ( $eff_name, $value )  = ( $1, $2 );
			my $survivehead = ( defined $3 and $3 =~ m/,\s*S(urvives)?\s*/i )
			 ? ", \"survives\"" : '';
			my $survivebody = ( $survivehead ne '' )
			 ? ', 1' : '';
			  
			# 'if' line
			if ( $e =~ s/^\s+if\s+(\w+)\s*,\s*(.+?)\s*,\s*(\w+)\s*\n?// ) {
				my ( $type, $typename, $range ) = ( $1, $2, $3 );

				$buf .= "[effect_$s->{type}_$s->{lc_name}_$count]\n"
				 . "name    = \"$eff_name\"\n"
				 . "value   = $value\n"
				 . "reqs    =\n"
				 . "    { \"type\", \"name\", \"range\"$survivehead\n"
				 . "      \"$type\", \"$typename\", \"$range\"$survivebody\n";
				$count++;

				while ( $e =~ s/^\s*(if|and)\s+(\w+)\s*,\s*(.+?)\s*,\s*(\w+)\s*\n?// ) {
					$buf .= "      \"$2\", \"$3\", \"$4\"\n";
				}
				$buf .= "    }\n";
			}
			elsif ( $s->{type} eq 'effects' ) { # one line effect
				$buf .= "[effect_$s->{lc_name}_$count]\n"
				 . "name    = \"$eff_name\"\n"
				 . "value   = $value\n\n";
				$count++;
			}

			if ( $e =~ m/^\s*unless/ ) {
				$buf .= "nreqs   =\n    { \"type\", \"name\", \"range\"\n";
				while ( $e =~ s/^\s*(unless|and)\s+(\w+)\s*,\s*(.+?)\s*,\s*(\w+)\s*\n?//m ) {
					$buf .= "      \"$2\", \"$3\", \"$4\"\n";
				}
				$buf .= "    }\n";
			}
			$buf .= "\n";

			# any errors
			if ( $e =~ m/\w/ ) {
				die "Looks like a bad Effect for $s->{type} $s->{name} (from $s->{filename}) \n$head\n$effect\n";
			}
		}
	}
	return $buf;
}

# returns formatted reqs
sub getReqs {
	my $s = shift;

	if ( not exists $s->{fields}{reqs} or not defined $s->{fields}{reqs}
	 or $s->{fields}{reqs} =~ m/^\s*$/ or $s->{fields}{reqs} eq '""' ) {
		if ( $s->{type} eq 'buildings' ) {
			return "reqs =\n    { \"type\", \"name\", \"range\"\n\n    }\n"; 
		}
		elsif ( $s->{type} eq 'governments' or $s->{type} eq 'cities' ) {
			return "; no reqs\n";
		}
		else {
			die "sub getReqs call for $s->{type} $s->{name}";
		}
	}
	my $reqs = $s->{fields}{reqs}; 
	my $buf = "reqs = \n    { \"type\", \"name\", \"range\"\n";

	$reqs =~ s/^\s*(\w+)\s*,\s*(.+?)\s*,\s*(\w+)\s*$//m
	 or die "$s->{type} $s->{name} (from $s->{filename}) has bad reqs\n'$reqs'";
	$buf .= "      \"$1\", \"$2\", \"$3\"\n";

	while ( $reqs =~ s/^\s+and\s+(\w+)\s*,\s*(.+?)\s*,\s*(\w+)\s*$//m ) {
		$buf .= "      \"$1\", \"$2\", \"$3\"\n";
	}
	$buf .= "    }\n";

	return $buf;
}

sub getResource {
	my $s = shift;

	$s->{fields}{resource} =~ m/\s*(\w+)\s*,\s*(\w+|-)/
	 or die "$s->{type} $s->{name} (from $s->{filename}) has bad resource field\n$s->{fields}{resource}\n";
	my ( $main, $alt ) = ( $1, $2 );

	if ( $s->{type} eq 'governments' ) {
		return
		   "graphic              = \"gov.$main\"\n"
		 . 'graphic_alt          = "' . (( $alt eq '-' )? '-' : "gov.$alt") . "\"\n";
	}
	elsif ( $s->{type} eq 'units' ) {
		return
		   "graphic              = \"u.$main\"\n"
		 . 'graphic_alt          = "' . (( $alt eq '-' )? '-' : "u.$alt") . "\"\n"
		 . "sound_move           = \"m_$main\"\n"
		 . 'sound_move_alt       = "m_' . (( $alt eq '-' )? 'generic' : $alt) . "\"\n"
		 . "sound_fight          = \"f_$main\"\n"
		 . 'sound_fight_alt      = "f_' . (( $alt eq '-' )? 'generic' : $alt) . "\"\n";
	}

	elsif ( $s->{type} eq 'buildings' ) {
		my $sound = ( $s->{genus} eq 'GreatWonder' )? 'w' : 'b';
		return
		   "graphic              = \"b.$main\"\n"
		 . 'graphic_alt          = "' . (( $alt eq '-' )? '-' : "b.$alt") . "\"\n"
		 . "sound                = \"${sound}_$main\"\n"
		 . "sound_alt            = \"${sound}_" . (( $alt eq '-' )? 'generic' : $alt ) . "\"\n";
	}
	elsif ( $s->{type} eq 'cities' ) {
		# get citizens
		die "No 'citizens' field defined for city $s->{name} (from $s->{filename})\n"
		 unless ( exists $s->{fields}{citizens} and defined $s->{fields}{citizens} );
		die "Bad 'citizens' field defined for city $s->{name} (from $s->{filename})\n"
		 unless ( $s->{fields}{citizens} =~ m/(\w+)\s*,\s*(\w+)/ ); 
		my ( $cit, $citalt ) = ( $1, $2 );

		return
		   "graphic               = \"city.$main\"\n"
		 . 'graphic_alt           = "' . (( $alt eq '-' )? '-' : "city.$alt") . "\"\n"
		 . "citizens_graphic      = \"$cit\"\n"
		 . "citizens_graphic_alt  = \"$citalt\"\n";
	}
	else {
		die "Bad type '$s->{type}' (from $s->{filename}) in getResource";
	}
}

sub getRulerTitles {
	my $s = shift;

	$s->{fields}{resource} =~ m/\s*(\w+)\s*,\s*(\w+|-)/
	 or die "$s->{type} $s->{name} (from $s->{filename}) has bad ruler_titles field\n$s->{fields}{ruler_titles}\n";

	return 
	   "ruler_male_title      = _(\"$1\")\n"
	 . "ruler_female_title    = _(\"$2\")\n\n";
}

sub getHelpText {
	my $s = shift;
	my $buf = $s->{fields}{helptext};
	return "\n; no helptext yet\n\n"
	 if $buf =~ m/^\s*$/; 

	while ( $buf =~ m/(\<([$TAGS])\s+(.+?)\s*\>)/ ) {
		my ( $tag, $type, $options ) = ( $1, $tagz{$2}, $3 );
		# | separated
		if ( $options =~ m/\|/ ) {
			my $gotone = 0;
			for my $opt ( split /\s*\|\s*/, $options ) {
				if ( exists $got->{$type}{$opt} ) {
					$gotone = 1;
					$buf =~ s/\Q$tag/$opt/g;
					last;
				}
			}
			$buf =~ s/[^\.]*?\Q$tag\E.*?\.\s*/. /gs
			 unless $gotone;
		}
		else { # , separated
			my ( $tmp, $c ) = ( '', 0 );
			for my $opt ( split /\s*,\s*/, $options ) {
				if ( exists $got->{$type}{$opt} ) {
					$c++;
					$tmp = $opt         if $c == 1;
					$tmp .= " or $opt"  if $c == 2;
					$tmp = "$opt, $tmp" if $c >= 3;
				}
			}
			if ( $c ) {
				$buf =~ s/\Q$tag/$tmp/;
			}
			else {
				$buf =~ s/[^\.]*?\Q$tag\E.*?(\.\s*|$)//gs;
			}
		}
	}

	$buf =~ s/\s*\n/\n /g;
	$buf =~ s/\n/\\\n/g;
	$buf =~ s/(\n\s*\.\\)+\n\s*/\n\\n\\n\n/g;

	return "helptext              = _(\"\n$buf\\\n\")\n\n";
}

# combat = attack, defense, hitpoints, firepower (for units only)
sub getUnitCombat {
	my $s = shift;

	$s->{fields}{combat} =~ m/\s*(\d+)\s*,\s*(\d+)\s*,\s*(\d+)\s*,\s*(\d+)/
	 or die "$s->{type} $s->{name} (from $s->{filename}) has bad combat field\n$s->{fields}{combat}\n";

	return
	   "attack               = $1\n"	
	 . "defense              = $2\n"	
	 . "hitpoints            = $3\n"	
	 . "firepower            = $4\n";
}

sub getUnitMove {
	my $s = shift;

	$s->{fields}{move} =~ m/\s*(\d+)(\s*,\s*(\d+))?/
	 or die "$s->{type} $s->{name} (from $s->{filename}) has bad move field\n$s->{fields}{move}\n";
	my $f = ( defined $3 ) ? $3 : 0;

	return
	   "move_rate            = $1\n"
	 . "fuel                 = $f\n"; 
}

sub getUnitUpkeep {
	my $s = shift;

	$s->{fields}{upkeep} =~ m/\s*(\d+)\s*,\s*(\d+)\s*,\s*(\d+)\s*,\s*(\d+)/
	 or die "$s->{type} $s->{name} (from $s->{filename}) has bad upkeep field\n$s->{fields}{upkeep}\n";

	return
	   "uk_happy             = $1\n"	
	 . "uk_shield            = $2\n"	
	 . "uk_food              = $3\n"	
	 . "uk_gold              = $4\n";
}

# put quotes around words, and add commas between, for flags and roles
sub layoutWords {
	my ( $s, $field ) = @_;

	my $buf;
	for ( split /\s*,\s*/, $s->{fields}{$field} ) {
		$buf .= "\"$_\", ";
	}
	$buf =~ s/, $//;
	$buf ||= '""';
	return "$field              = $buf\n";
}

# general field getter
# types = (s)tring, (d)igits, (n)ame
# flags = (r)equired, can be (n)one, can be (c)ancelled, default to (z)ero, can be replaced with '-'
# realname - not required
sub getField {
	my ( $s, $field, $type, $flag, $realname ) = @_;
	my $rfield = ( defined $realname )? $realname : $field;

	my $val = '';
	if ( not defined $s->{fields}{$field} or $s->{fields}{$field} eq '' ) {
		if    ( $flag eq 'r' ) { die "$s->{type} $s->{name} (from $s->{filename}) has empty field '$field'\n" }
		elsif ( $flag eq 'c' ) { return '' }
		elsif ( $flag eq '-' ) { return "$rfield   = \"-\"\n" }
		elsif ( $flag eq 'n' ) { $val = 'None' } # assumes $type = string
		elsif ( $flag eq 'z' ) { $val = 0      } # assumes $type = digit
		else  { die "Bad Flag '$flag' for $s->{name} (from $s->{filename}):$field\n" }
	}
	elsif ( $type eq 'n' ) {
		return "name                 = _(\"$s->{name}\")\n";
	}
	elsif ( $type eq 's' ) {
		die "$s->{type} $s->{name} (from $s->{filename}) has field $field that wants a string, instead has '$s->{fields}{$field}'\n"
		 if $s->{fields}{$field} =~ m/^\d+$/;
		return "$rfield             = \"$s->{fields}{$field}\"\n";
	}
	elsif ( $type eq 'd' ) {
		die "$s->{type} $s->{name} (from $s->{filename}) has field $field that wants a number, instead has '$s->{fields}{$field}'\n"
		 unless $s->{fields}{$field} =~ m/^\d+$/;
		return "$rfield            = $s->{fields}{$field}\n";
	}
	die "Unknown type '$type' in $s->{name}: $field (from $s->{filename})";
}

1;

